# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

__version__ = '1.1.14'

SETUPTOOLS_REQUIREMENT = 'setuptools>=2.2,<20.11'
WHEEL_REQUIREMENT = 'wheel>=0.24.0,<0.30.0'
